/*
  GPS Mini Navigator    for navigating predefined ways & extended gps logging

  Copyright (C) 2008  Matthias Larisch

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation; either version 2 of the License, or (at your option) any later
  version.
  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  You should have received a copy of the GNU General Public License along with
  this program; if not, see <http://www.gnu.org/licenses/>.



 Extended GPS Logger - a project 2008 by Matthias Larisch
 Idea: beginning of 2008
 started to plan: may 2008
 HW built: july 2008
 coding start: 14th july 2008
 features finished & small bugfix: 31th august 2008

UART: HW to GPS (38400 bps)
      SW (PB0 - RXD PB1 - TXD) (38400 bps to Bluetooth)

SWITCH: PD5, PD7 (int. pullups, sw to gnd)

Powermanagement: FET switchable peripherals: PD3, PD4 high is OFF

SPI: SD card, SS is PB2 (PB3,PB4,PB5)

By Ports:
PORTB 0: SU RX 1: SU TX 2: SS  3: MOSI 4: MISO 5: SCK  6: XTAL1 7: XTAL2
PORTC 0: LED   1: LED_C 2: LED 3: LED  4: LED  5:      6:       7:
PORTD 0: RXD   1: TXD   2:     3: FET1 4: FET2 5: SW A 6:  7: SW B


Uart communication:
GPS: 38400 Sirf binary protocol

BlueTooth: Startup: C command mode. Switchable to direct GPS
Switch back: Restart Bluetooth Module (external switch)

*/
#include "egl.h"
#include "suart.h"
#include "suart_commands.h"
#include "mmc.h"
#include "filesystem_simple.h"
#include "gps_uart.h"



//volatile struct status_t status;
volatile struct power_status_t power_status;

uint8_t gps_log_interval;
uint8_t gps_log_distance;

gps_coords nav_to;
uint8_t navi_file = 0xFF;
uint16_t navi_entry;

volatile uint8_t bluetooth_on_cnt;
#define bluetooth_on_cnt_max 25		// Enable/Disable SW Uart n*Timer0 Interrupts after
									// RX going low (-> off) or high (->on)

// to do: led vars: 2 bits per led, 00 = off, 01 = low, 1x = on
volatile uint8_t led;
// 7 0- 6 0+ 5 3+ 4 3- 3 2+ 2 2- 1 4- 0 4+

const uint8_t costable[] PROGMEM = {
100,100,100,100,100,100,99,99,99,99,98,		// cos 0 - cos 10
98,98,97,97,97,96,96,95,95,94,				// cos 11 - cos 20
93,93,92,91,91,90,89,88,87,87,				// cos 21 - cos 30
86,85,84,83,82,81,80,79,78,77,				// cos 31 - cos 40
75,74,73,72,71,69,68,67,66,64,				// cos 41 - cos 50
63,62,60,59,57,56,54,53,52,50,				// cos 51 - cos 60
48,47,45,44,42,41,39,37,36,34,				// cos 61 - cos 70
33,31,29,28,26,24,22,21,19,17,				// cos 71 - cos 80
16,14,12,10,9,7,5,3,2,0};					// cos 81 - cos 90

static uint8_t my_cos(uint8_t a)		// simplified cosinus function, returns only positive values
{
  //while(a > 90) a-=90;				// not needed, worst case we get wrong cos value
  return pgm_read_byte(&costable[a]);
}

static uint8_t my_acos(uint8_t a)		// simplified acos function
{
  if(a >= 100) return 2;		// errorcheck & better results in one line :-)
  uint8_t temp = 0;
  while(pgm_read_byte(&costable[temp]) > a) temp++;
  return temp;
}

// returned distance is about +/- 1% correct
// flash usage: 418 bytes including inlined cos/acos
// add 92 bytes for costable (->510 bytes)
uint32_t calculate_distance (gps_coords *a, gps_coords *b)
{
  uint8_t latstate = 0;
  uint32_t lat_distance;
  if(a->lat > b->lat)
  {
    latstate = 1;
	lat_distance = a->lat - b-> lat;
  }
  else lat_distance = b->lat - a->lat;
   

  //uint32_t lat_distance = mul32_16u(div32u(labs(a->lat - b->lat),100),11112);	// * 10^7 * 10^-2 *10^-1= *10^4 meters
  lat_distance = mul32_16u(div32u(lat_distance,100),11112);	// * 10^7 * 10^-2 *10^-1= *10^4 meters
  //uint32_t lat_distance = mul32_16u(div32u(abs(a->lat - b->lat),100),284);	// * 10^7 * 10^-2 *10^-1= *10^4 meters
  lat_distance = div32u(lat_distance, 10000);
  uint16_t distance_cache = lat_distance;			
  lat_distance = mul32_16u(lat_distance, lat_distance);		// square of lat_distance

  uint8_t lonstate = 0;
  uint32_t lon_distance;
  if(a->lon > b->lon)
  {
    lonstate = 1;
	lon_distance = a->lon - b->lon;
  }
  else lon_distance = b->lon - a->lon;
  //uint32_t lon_distance = mul32_16u(div32u(labs(a->lon - b->lon),1000), my_cos(div32u(a->lat,10000000)));  //->10^6  // simplification: a->lat instead of a->lat+b->lat/2
  //lon_distance = mul32_16u(div32u(lon_distance,1000), my_cos(div32u(a->lat,10000000)));  //->10^6  // simplification: a->lat instead of a->lat+b->lat/2
  /* square(((lon_diff / 1000) * cos_val * 1113) / 10000) = lon_diff * 111,3 * cos(angle)*/
  lon_distance = mul32_16u(div32u(lon_distance, 1000), my_cos(div32u(labs(a->lat),10000000)));  //->10^6  // simplification: a->lat instead of a->lat+b->lat/2
  //lon_distance = mul32_16u(div32u(lon_distance, 1000), my_cos(div16u(abs(a->lat >> 16),153)));  //->10^6  // simplification: a->lat instead of a->lat+b->lat/2
  lon_distance = div32u(mul32_16u(lon_distance, 1113),10000); // ->meters
  lon_distance = isqrt32(mul32_16u(lon_distance, lon_distance) + lat_distance);
  uint8_t temp_angle = (my_acos(div32u(mul32_16u(distance_cache,100),lon_distance)) * 100) >> 8;	// 0-90 deg. angle
 
  
  
  if(latstate)
  {
    if(lonstate) temp_angle +=70;
	else temp_angle = 70 - temp_angle;
  }
  else if(lonstate)
    temp_angle = 140 - temp_angle;
  
  //return (lon_distance) | ((uint32_t)temp_angle << 16);
  asm volatile ("clr %D0"		"\n\t"
  				"mov %C0, %2"   "\n\t"
                 : "+r" (lon_distance)
				 : "0" (lon_distance), "r" (temp_angle));
  return lon_distance;

  //uint32_t lon_distance = (uint32_t)abs(a->lon - b->lon) * (uint32_t)((uint32_t)my_cos(div32u((a->lat + b->lat)>>1,10000000))*1113UL);
  //32 bit square root -> single factor max. 16 bit
  //single length max. 65535 meters (5882352 * 10^-7 degrees lat)
  // lat: 11112 *10^-1 *10^7 = *10^6 meters = micrometers 

}

static void init( void )
{
  MCUSR &= ~(1 << WDRF);
  wdt_disable();
  ADCSRA = (1 << ADEN) | (1 << ADSC) | (1 << ADPS1) | (1 << ADPS2);
  PRR = (1 << PRTWI) | (1 << PRTIM2) | (1 << PRSPI) | (1 << PRUSART0);	// disable unused IO
  DDRB = 0xFE;  // PortB as Output and zero, SRX is Input
  PORTB = 0;

  DDRC = 0xFF;  // PortC as Output and zero
  PORTC = 0;

  DDRD = (0 << PD7) | (1 << PD6) | (0 << PD5) | (1 << PD4)
         | (1 << PD3) | (0 << PD2) | (0 << PD1) | (0 << PD0);
  PORTD = (1 << PD3) | (1 << PD4) | (1 << PD5) | (1 << PD7); // fet off, switch pullups
  //suart_init();
  TCCR0B = (1 << CS00) | (1 << CS01);		// Timer 0 Fio/64 for LEDs
  TIMSK0 = (1 << TOIE0);					// Overflow interrupt every 2,048 ms
  /*
  CLKPR = (1 << CLKPCE);
  CLKPR = (1 << CLKPS2) | (1 << CLKPS1);	// System Clock Prescaler / 64
  */

  /* taken from suart_init() */

  //STXDDR |= 1<<STX;			// TX output    // already done above
  sei();
  led = 0xAA;
}

ISR(TIMER0_OVF_vect)			// ~97 cycles max in normal uart mode
/* warning: interrupt response time for software uart has to be below bittime (207 cycles)!
that means that ALL interrupts together MUST NOT exceed 207 cycles.
To be exact, OC1A interrupt must be called 2 times in ~300 cycles so we have about 250 cycles
MAX.*/
{
  uint8_t temp = 0;
  uint8_t temp_led = led;
  uint8_t temp2;
  temp2 = SRXPIN & (1 << SRX);
  if(power_status.sw_uart_on)		
  {    
    if(temp2) bluetooth_on_cnt = 0;
	else bluetooth_on_cnt++;
  }
  else
  {
    static uint8_t ct0, ct1, rpt;
    static uint8_t second_prescaler = 245;
    if(status & (1 << st_led))
    {
      if(second_prescaler-- == 0)
	  {
	    second_prescaler = 245;			// here we have ~1 sec clock
	    status |= (1 << st_second);
      }
      if(status & (1 << st_hold))
	    return;						// RETURN if hold!
	  temp = key_state ^ ~KEY_PIN;		// key changed ?		// key debouncing every 4,096 ms
      ct0 = ~( ct0 & temp );			// reset or count ct0
      ct1 = ct0 ^ (ct1 & temp);		// reset or count ct1
      temp &= ct0 & ct1;			// count until roll over ?
      key_state ^= temp;			// then toggle debounced state
      key_press |= key_state & temp;		// 0->1: key press detect

      if( (key_state & REPEAT_MASK) == 0 )	// check repeat function
        rpt = REPEAT_START;		// start delay
      if( --rpt == 0 ){
        rpt = REPEAT_NEXT;			// repeat delay
        key_rpt |= key_state & REPEAT_MASK;
      }    
	
	  temp = 0x1F;						// all + common high
	  if(temp_led & 1) temp &= ~(1 << 4);
	  if(temp_led & (1 << 3)) temp &= ~(1 << 2);
	  if(temp_led & (1 << 5)) temp &= ~(1 << 3);
	  if(temp_led & (1 << 6)) temp &= ~1;
	/*
	temp = (led & (1 << 0)) << 4;
	temp |= (led & (1 << 3)) << 2;
	temp |= (led & (1 << 5)) << 3;
	temp |= (led & (1 << 6)) << 0;
	*/
    }
    else
    {
      if(temp_led & (1 << 1)) temp |= 1 << 4;
	  if(temp_led & (1 << 2)) temp |= 1 << 2;
	  if(temp_led & (1 << 4)) temp |= 1 << 3;
	  if(temp_led & (1 << 7)) temp |= 1;
	
	/*temp = (!(led & (1 << 1))) << 4;
	temp |= (!(led & (1 << 2))) << 2;
	temp |= (!(led & (1 << 4))) << 3;
	temp |= (!(led & (1 << 7))) << 0;
	*/
    }
    PORTC = temp;
    //status.led=1-status.led;
    if(status & (1 << st_led)) status &= ~(1 << st_led);
    else status |= (1 << st_led);
    if(temp2) bluetooth_on_cnt++;	// sw uart receive routine also resets counter
	else bluetooth_on_cnt = 0;		// on each received byte, so small count is sufficient
  }
  if(bluetooth_on_cnt >= bluetooth_on_cnt_max)
  {
    bluetooth_on_cnt = 0;
	//power_status.sw_uart_on = temp2;
	if(temp2) status |= (1 << st_start_sw_uart);
	else suart_deinit();
	
  }

  

}

// todo: dont show leading zeros
uint8_t decimal_display_table_p[] PROGMEM = {0x83, 1, 2, 4, 8, 16, 32, 64, 128, 0xC0};
void decimal_display(uint32_t number)			// displays a decimal number on leds (9 is led 8 + led 7, 0 is led 8 1 and 2)
{
  static uint8_t state;				// state = 0 -> millions 1-> 100 thousands 2-> 10 thousands 3-> thousands 4-> hundreds 5-> tens 6-> ones
  static uint32_t cached_number;
  if(number)
  {
    cached_number = number;
	status |= (1 << st_decimal_display);
	state = 0;
  }
  uint8_t temp = 0;
  uint32_t sub = 1000000000;
  uint8_t temp_state;
  uint32_t work_number = cached_number;
  //uint8_t not_done = 1;
  if(state == 11 || work_number == 0) state = 0;
  else if(state > 9) led = 0xFF;
  else if(state <= 9)
  {
    //while(not_done)
    {
      for(temp_state = 0; temp_state <= state; temp_state++)
      {
        temp = 0;
	    while(work_number >= sub)
	    {
	      work_number -= sub;
	      temp++;
        }
	    sub = div32u(sub, 10);
		//sub/=10;
      }
      /*if(temp == 0) temp = 0x83;
      else if(temp >= 1 && temp <= 8) temp = 1 << (temp-1);
      else if(temp == 9) temp = 0xC0;
   else if(temp == 1) temp = 0x01;
  else if(temp == 2) temp = 0x02;
  else if(temp == 3) temp = 0x04;
  else if(temp == 4) temp = 0x08;
  else if(temp == 5) temp = 0x10;
  else if(temp == 6) temp = 0x20;
  else if(temp == 7) temp = 0x40;
  else if(temp == 8) temp = 0x80;
  else if(temp == 9) temp = 0xB0;
  
      if(temp > 0 || work_number & (1UL << 31))
	  {
	    not_done = 0;
		cached_number |= (1UL << 31);
      }
	  else state++;*/


    }
	led = pgm_read_byte(&decimal_display_table_p[temp]);
	//led = temp;
  }
  state++;
}

//const uint8_t led_degree[] PROGMEM = {127, 139, 0, 13, 5, 21, 19, 37, 30, 64, 52, 88, 76, 110, 103, 121, 119, 135};
const uint8_t led_degree[] PROGMEM = {119, 135, 103, 121, 76, 110, 52, 88, 30, 64, 19, 37, 5, 21, 0, 13, 127, 140};
// 140 should not be neccessary but i don't want to look at every detail
// the table contains interval values for every led beginning at LED 7 counterclockwise to
// LED zero which has two intervals (< 360 and greater 0)
// displays course with leds, absolute = 1 -> direct, absolute = 0: add current heading
uint8_t display_course(uint8_t course, uint8_t direct)
{
  if(!direct)
  {
    /*uint16_t temp = course + current_gps_heading;
	if(temp > 139) temp -= 140;
	course = temp;
	*/
	int16_t temp = course - current_gps_heading;
	if(temp < 0) temp += 140;
	course = temp;
  }

  // led display: 
  uint8_t led_t = 0;
  
  
  uint8_t i = 0x80, j = 0;
  uint8_t temp, temp2;
  do
  {
    temp = pgm_read_byte(&led_degree[j]);
	temp2 = pgm_read_byte(&led_degree[j+1]);
	j+=2;
	if(temp <= course && temp2 >= course)
	  led_t |= i;
    if(i > 1) i = i >> 1;
    //j --;
  } while(j <= 17);
  return led_t;
}

// ADC conversion of 1.1 V internal bandgap voltage with AVcc Reference
// values for my own hardware:
// 3263 mV -> 68
// 3088 mV -> 87
// 2880 mV -> 115
// 2792 mV -> 126
// I don't use ADLAR because high byte is always 1 in this voltage range.
// higher voltages than 3.3 V are not possible because of the 3.3 V LDO.
// Measures for my 1100 mAh Liion:
// 75 mA load. voltage measures BEFORE regulator:
// t=0    U=3,50 V
// t=18 m U=3,27 V
// t=21 m U=3,20 V
// t=23 m U=3,10 V
// t=24 m U=3,00 V
// t=26 m U=2,80 V
// this was continuous load by GPS + C. After a few hours, the battery was at 3,1 V
// without load again. For correct SD-Card access, we need 2,7 V behind the regulator.
// so we have to disable enabling of peripherie at about 3,0 V (~95) and force disable
// (if enabled) at 2,7 V (~130). C start from shutdown can be allowed until 135.

uint8_t __attribute__((noinline)) get_power()
{
  ADMUX = (1 << REFS0) | (1 << MUX3) | (1 << MUX2) | (1 << MUX1);
  ADCSRA = (1 << ADEN) | (1 << ADSC) | (1 << ADPS1) | (1 << ADPS2);
  while(ADCSRA & (1 << ADSC));
  ADCSRA = (1 << ADEN) | (1 << ADSC) | (1 << ADPS1) | (1 << ADPS2);
  while(ADCSRA & (1 << ADSC));
  return ADC;
}


ISR(PCINT2_vect)		// right key press detect
{
  #define RIGHT_NOT_PRESSED (KEY_PIN & (1 << KEY1))
  #define RIGHT_PRESSED 	((KEY_PIN & (1 << KEY1)) == 0)
  #define LEFT_NOT_PRESSED	(KEY_PIN & (1 << KEY0))
  #define LEFT_PRESSED		((KEY_PIN & (1 << KEY0)) == 0)
  
  if(RIGHT_PRESSED && LEFT_NOT_PRESSED)
  {
    _delay_ms(1000);
	if(RIGHT_NOT_PRESSED && LEFT_NOT_PRESSED)
	{
	  _delay_ms(500);
	  if(RIGHT_PRESSED && LEFT_PRESSED)
	  {
	    status &= ~(1 << st_hold);
		PCICR = 0;
      }
    }
  }
  
  PCIFR |= (1 << PCIF2);
  return;
}
/*
  if((KEY_PIN & (1 << KEY1)) || (KEY_PIN & (1 << KEY0))==0)	// right key pressed, left key released.
  return;
  _delay_ms(700);											// after 700 ms, right key must be released
  if((KEY_PIN & (1 << KEY1))==0 || (KEY_PIN & (1 << KEY0))==0)
  return;
  _delay_ms(400);											// 400 ms later, we want both keys pressed
  if((KEY_PIN & (1 << KEY1))==0 || (KEY_PIN & (1 << KEY0))==0)
  {
    status &= ~(1 << st_hold);
    PCICR = 0;				// disable interrupt on success
    PCIFR |= (1 << PCIF2);
  }
  return;
*/
/*
summary:
t			l	r
0-699  		O	X
700-1099	O	O
1100		X	X
*/		
//}

void shutdown()
{
  PCICR = (1 << PCIE2);
  PCMSK2 = (1 << PCINT21);

  if(FETPORT&(1 << GPS_FET))
  //GPS OFF (1)
  {
    SMCR = (1 << SM1) | (1 << SE);
	mmc_shutdown();
    ADCSRA = 0;
    uint8_t temp = PRR;
    PRR = 0xFF;
	status |= (1 << st_hold);
    while (status & (1 << st_hold))
	  asm volatile ("sleep");				// Power down mode, only right key allows wakeup
	SMCR = 0;							// 33 - 35 A total consumption for my navi board in this state
	PRR = temp;
	mmc_init();
	ADCSRA = (1 << ADEN) | (1 << ADSC) | (1 << ADPS1) | (1 << ADPS2);
  }
  else
  // GPS ON -> no shutdown!
  {
    status |= (1 << st_hold);			// disable key and led handling -> hold
  }
  
}

int __attribute__((OS_main)) main()
{
init();
mmc_init();
uint16_t seconds = 0;
gps_coords last_log;
while(1)
{
  
  if(status & (1 << st_start_sw_uart))
  {
    suart_init();
    status &= ~(1 << st_start_sw_uart);
  }
  if(power_status.sw_uart_on)  suart_receive();		// ensure that this function is called at least 1 time every 1900 cycles
  else
  {
    ui_do();										// all key/menu handling, everything...
  
    if(status & (1 << st_second))
    {
      decimal_display(0);
	  status &= ~(1 << st_second);
	  if((FETPORT&(1 << GPS_FET)) == 0)
	  {
	    if(get_power() >= PWR_GPS_RUNNING_LIMIT) gps_disable();		// GPS Disable on battery low
      }
	  seconds++;
    }
    if(status & (1 << st_gps_log))
    {
      if(status & (1 << st_gps_log_entry_updated))
	  {
	    uint8_t update_log = 0;
	    if(gps_log_distance)
	    {
	      uint16_t temp_dist = calculate_distance((gps_coords *)&gps_log_entry.coords, &last_log);
		  if((gps_log_distance == 1 && temp_dist >= 50)  ||
		     (gps_log_distance == 2 && temp_dist >= 400) ||
		     (gps_log_distance == 3 && temp_dist >= 2000)) update_log = 1;
        }

	    if((gps_log_interval == 1 && seconds == 16) ||
	       (gps_log_interval == 2 && seconds == 128) ||
	       (gps_log_interval == 3 && seconds == 1024) ||
	       (gps_log_interval == 0)) update_log = 1;
	  
	    if(update_log)
	    {
	      last_log.lat = gps_log_entry.lat;
		  last_log.lon = gps_log_entry.lon;
		  write_file((uint8_t *)&gps_log_entry,20);
	      status &= ~(1 << st_gps_log_entry_updated);
		  seconds = 0;
        }
  	  }
    }
  }
}
}
