#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#define F_CPU 8000000UL
#include <util/delay.h>
#include <stdlib.h>


#define XTAL 8000000UL
#define SW_BAUD 38400
#define HW_BAUD 38400

#define FETPORT PORTD
#define FET1 PD3
#define FET2 PD4

#define GPS_FET FET1
#define SD_FET FET2

extern volatile uint8_t bluetooth_on_cnt;
extern volatile uint8_t led;

/*struct status_t
{
unsigned led:1;
unsigned start_sw_uart:1;
unsigned gps_log_entry_updated:1;
unsigned gps_log:1;
unsigned second:1;
unsigned decimal_display:1;
unsigned gps_always_on;
};*/

#define status GPIOR0		// where to safe often required status flags? recommended to use low IO reg for Flash saving

#define st_led 0
#define st_start_sw_uart 1
#define st_gps_log_entry_updated 2
#define st_hold 3
#define st_second 4
#define st_decimal_display 5
#define st_gps_always_on 6
#define st_gps_log 7				// this should be 7 for space-saving in keys.c

struct power_status_t
{
unsigned sw_uart_on:1;
//unsigned fet1_on:1;		// FET 1 -> PD3 -> GPS
//unsigned fet2_on:1;		// FET 2 -> PD4 -> SD
};

typedef union
  {
    struct
    {
	  int32_t lat;
      int32_t lon;
      int32_t alt;
    };
	uint8_t coords[12];
  } gps_coords;


//extern volatile struct status_t status;
extern volatile struct power_status_t power_status;

#define KEY_PIN		PIND
#define KEY0		7
#define KEY1		5

#define REPEAT_MASK	(1<<KEY0^1<<KEY1)	// repeat: key1, key2
#define REPEAT_START	125		// after 500ms
#define REPEAT_NEXT	70		// every 200ms

extern volatile uint8_t key_state;				// debounced and inverted key state:
										// bit = 1: key pressed
extern volatile uint8_t key_press;				// key press detect
extern volatile uint8_t key_rpt;				// key long press and repeat

extern uint32_t cached_number;

extern uint8_t gps_log_interval;
#define GPS_LOG_INTERVALS 4

extern uint8_t gps_log_distance;
#define GPS_LOG_DISTANCES 4

#define DISTANCE_AUTO_NEXT_WAYPOINT 20			// when enabled navigate to next waypoint if current waypoint
												// is nearer than X meters

extern gps_coords nav_to;
extern uint8_t navi_file;
extern uint16_t navi_entry;

void decimal_display(uint32_t);
uint8_t display_course(uint8_t, uint8_t);
void shutdown();


uint32_t calculate_distance (gps_coords *, gps_coords *);

extern void ui_do();
extern uint8_t menu;	// from ui.c

extern uint16_t isqrt32 (uint32_t);
extern uint32_t div32u (uint32_t, uint32_t);
extern uint16_t div16u (uint16_t, uint16_t);
extern uint32_t mul32_16u (uint32_t, uint16_t);
extern uint8_t get_power(void);

#define PWR_GPS_ON_LIMIT 95
#define PWR_GPS_RUNNING_LIMIT 130
