/*
  GPS Mini Navigator    for navigating predefined ways & extended gps logging

  Copyright (C) 2008  Matthias Larisch

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation; either version 2 of the License, or (at your option) any later
  version.
  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  You should have received a copy of the GNU General Public License along with
  this program; if not, see <http://www.gnu.org/licenses/>.
  

  file description: GPS handling with Sirf Binary Protocol

*/


#include "egl.h"
#include "gps_uart.h"

char pgmInitSirfMode[] PROGMEM = "\x1b" "$PSRF100,0,38400,8,1,0*3C\r\n";
//char pgmSirfMessageStart[] PROGMEM = "\xA0" "\xA2" "\0";

volatile gps_log_entry_t gps_log_entry;
volatile uint8_t current_gps_heading;		//( 2.57 deg. steps, starting from north clockwise)


void gps_enable (uint8_t rxint)
{
  PRR &= ~(1 << PRUSART0);
  UBRR0 = GPS_UBRRVAL;
  UCSR0C = (1 << UCSZ01) | (1 << UCSZ00);
  FETPORT &= ~(1 << GPS_FET);
  uint8_t temp = (1 << RXEN0) | (1 << TXEN0) | (1 << RXCIE0);
  if(rxint == 0) temp = (1 << RXEN0) | (1 << TXEN0);
  UCSR0B = temp;
  _delay_ms(100);
  gps_transmit_p(pgmInitSirfMode);
  uint8_t buffer[8];
  uint8_t i;
  for(i=0;i<=7;i++)
    buffer[i] = 0;
  buffer[0] = 0xA6;
  buffer[2] = 0x02;
  send_sirf_message(buffer,8);
  buffer[2] = 0x04;
  send_sirf_message(buffer,8);
  buffer[2] = 0x07;
  send_sirf_message(buffer,8);
  buffer[2] = 0x09;
  send_sirf_message(buffer,8);
  buffer[2] = 0x1B;
  send_sirf_message(buffer,8);
  /* $PSRF100,0,38400,8,1,0*3C<CR><LF> */ // Switch to SIRF mode
  /* Set message rate:
  A0 A2 00 08
  A6 00 02 05 00 00 00 00		MID;Send now?;MID;Rate
  00 AD B0 B3

  Messages: 0x1B (DGPS Status)
  0x07 (Clock Status)
  0x09 (Throughput)
  0x02 (measured nav data)
  0x04 (measured tracking data)
  0x29 (41) is what we want
  */
}

void gps_disable (void)
{
  UCSR0B = 0;
  FETPORT |= (1 << GPS_FET);
  //power_status.GPS_FET_PWR = 0;
  PRR |= (1 << PRUSART0);
}

void __attribute__((noinline)) gps_transmit  (uint8_t byte) 
{
  while (! (UCSR0A & (1 << UDRE0)));
  UDR0 = byte;
}

void __attribute__((noinline)) gps_transmit_p (PGM_P address)
{
  uint8_t count = pgm_read_byte(address++);
  while(count--)
  {
    gps_transmit(pgm_read_byte(address++));
  }
} 

void send_sirf_message(uint8_t *data, uint8_t length)
{
  gps_transmit(0xA0);
  gps_transmit(0xA2);
  gps_transmit(0x00);
  
  //gps_transmit_p (pgmSirfMessageStart);		// Saves 2 bytes flash

  gps_transmit(length);
  uint16_t crc = 0;
  while(length--)
  {
    gps_transmit(*data);
	crc += *data++;
  }
  gps_transmit((crc>>8)&0x7F);
  gps_transmit(crc&0xFF);
  gps_transmit(0xB0);
  gps_transmit(0xB3);
}


/* this ISR needs some size optimation, currently 686 bytes! */
ISR(USART_RX_vect)
{
  static uint8_t bytenum = 0;
  static gps_log_entry_t int_gps_log;
  static uint8_t temporary_store;
  static uint16_t checksum = 0;

  uint8_t uart_byte = UDR0;
  if(uart_byte != 0xA0 && bytenum == 0)
    return;
  if(uart_byte != 0xA2 && bytenum == 1)
  {
    bytenum = 0;
    return;
  }
  if(uart_byte != 0x00 && bytenum == 2)
  {
    bytenum = 0;
    return;
  }
  if(uart_byte != 0x5B && bytenum == 3)
  {
    bytenum = 0;
    return;
  }
  if(uart_byte != 0x29 && bytenum == 4)
  {
    bytenum = 0;
    return;
  }
  if(bytenum == 8)
  {
    int_gps_log.sog = uart_byte & 0x07;			// nav type [ fix type ]
	uint8_t temp2 = 0;
	if(uart_byte & 0x08)	temp2 = 0x80;
	else int_gps_log.hdop = temp2;				// nav type [ trickle power? ]
  }
  /* this can be discarded! we only support 6 bit year till 2072...
  if(bytenum == 15)								// year high byte
  {
    temporary_store = uart_byte - 7;			// manual subtraction of 2008
  }
  */
  if(bytenum == 16)
  {
    int_gps_log.date.s.year = (uart_byte - 216) & 0x3F;
	  // this calculates year 2008 = 0 ... 2072 = 0x3F
  }
  if(bytenum == 17)
  {
    int_gps_log.date.s.month = uart_byte & 0x0F;
  }
  if(bytenum == 18)
  {
    int_gps_log.date.s.day = uart_byte & 0x1F;
  }
  if(bytenum == 19)
  {
    int_gps_log.date.s.hour = uart_byte & 0x1F;
  }
  if(bytenum == 20)
  {
    int_gps_log.date.s.minute = uart_byte & 0x3F;
  }
  if(bytenum == 21)		// millisecond high byte
  {
    temporary_store = uart_byte;
  }
  if(bytenum == 22)
  {
    uint16_t seconds = (temporary_store << 8) | uart_byte;
	seconds = div16u(seconds, 1000);						// MAY FIX ME IF DIVISION IS ONLY NEEDED FOR THIS
	//seconds /= 1000;
	int_gps_log.date.s.second = seconds & 0x3F;
  }
  if(bytenum >= 27 && bytenum <= 34)			// lat / lon
  {
    uint8_t temp = bytenum - 27;
    if(temp > 3) temp = 11 - temp;
	else temp = 3 - temp;							// correct byte order
	int_gps_log.coords[temp] = uart_byte;
  }
  if(bytenum >= 39 && bytenum <= 42)
  {
	uint8_t temp = bytenum - 39;
	temp = 11 - temp;
	int_gps_log.coords[temp] = uart_byte;
  }
  if(bytenum == 44)								// sog high byte
  {
    temporary_store = uart_byte;
  }
  if(bytenum == 45)
  {
    int_gps_log.sog |= ((uint16_t)temporary_store << 8) | (uart_byte & 0xF8);
  }
  if(bytenum == 46)
  {
    current_gps_heading = uart_byte;
  }
  if(bytenum == 92)
  {
    int_gps_log.sv = uart_byte;
  }
  if(bytenum == 93)
  {
    int_gps_log.hdop |= uart_byte & 0x7F;
  }    
  if(bytenum == 4) checksum = 0;
  if(bytenum >= 4 && bytenum < 95) checksum += uart_byte;
  if(bytenum == 95 && ((checksum >> 8) & 0x7F) != uart_byte)
  {
    bytenum = 0;
    return;
  }
  if(bytenum == 96 && (checksum & 0xFF) != uart_byte)
  {
    bytenum = 0;
	return;
  }
  if(bytenum == 97 && uart_byte != 0xB0)
  {
    bytenum = 0;
	return;
  }
  if(bytenum == 98 && uart_byte == 0xB3)
  {
    gps_log_entry = int_gps_log;
	status |= (1 << st_gps_log_entry_updated);
	bytenum = 0;
	return;
  }

  bytenum++;

}
