/*#######################################################################################
Connect AVR to MMC/SD 

Copyright (C) 2004 Ulrich Radig
-> Veraenderungen Juli bis August 2008 von Matthias Larisch

Bei Fragen und Verbesserungen wendet euch per EMail an

mail@ulrichradig.de

oder im Forum meiner Web Page : www.ulrichradig.de

Dieses Programm ist freie Software. Sie knnen es unter den Bedingungen der 
GNU General Public License, wie von der Free Software Foundation verffentlicht, 
weitergeben und/oder modifizieren, entweder gem Version 2 der Lizenz oder 
(nach Ihrer Option) jeder spteren Version. 

Die Verffentlichung dieses Programms erfolgt in der Hoffnung, 
da es Ihnen von Nutzen sein wird, aber OHNE IRGENDEINE GARANTIE, 
sogar ohne die implizite Garantie der MARKTREIFE oder der VERWENDBARKEIT 
FR EINEN BESTIMMTEN ZWECK. Details finden Sie in der GNU General Public License. 

Sie sollten eine Kopie der GNU General Public License zusammen mit diesem 
Programm erhalten haben. 
Falls nicht, schreiben Sie an die Free Software Foundation, 
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA. 
#######################################################################################*/

#include "mmc.h"
#include "egl.h"
//#include "suart.h"
uint16_t mmc_sector_remaining;			// Zhler fr bereits gelesene Bytes eines Sektors

static uint8_t mmc_read_byte(void);
static void mmc_write_byte(uint8_t);
static uint8_t mmc_write_command(uint8_t *);

//############################################################################
//Routine zur Initialisierung der MMC/SD-Karte (SPI-MODE)
uint8_t mmc_init ()
//############################################################################
{
	uint16_t Timeout = 0;
	
	// Die Konfiguration wird in die Projekt-Initialisierung verlagert
	//Konfiguration des Ports an der die MMC/SD-Karte angeschlossen wurde
	/*MMC_Direction_REG &=~(1<<SPI_DI);			//Setzen von Pin MMC_DI auf Input
	MMC_Direction_REG |= (1<<SPI_Clock);		//Setzen von Pin MMC_Clock auf Output
	MMC_Direction_REG |= (1<<SPI_DO);			//Setzen von Pin MMC_DO auf Output
    MMC_Direction_REG |= (1<<SPI_SS);	
	*/

    FETPORT &= ~(1 << SD_FET);	// Powerhandling (Projektspezifisch)
	PRR &= ~(1 << PRSPI);  

	MMC_Direction_REG |= (1<<MMC_Chip_Select);	//Setzen von Pin MMC_Chip_Select auf Output
	MMC_Write |= (1<<MMC_Chip_Select);			//Setzt den Pin MMC_Chip_Select auf High Pegel

	for(uint8_t a=0;a<200;a++){
	nop();
	};		//Wartet eine kurze Zeit

	#if SPI_Mode
		//Aktiviren des SPI - Bus, Clock = Idel LOW
		//SPI Clock teilen durch 128
		SPCR = (1<<SPE)|(1<<MSTR)|(1<<SPR0)|(1<<SPR1); //Enable SPI, SPI in Master Mode	
		SPSR = (0<<SPI2X);
	#endif
	
	//Initialisiere MMC/SD-Karte in den SPI-Mode
	for (uint8_t b = 0;b<0x0f;b++) //Sendet min 74+ Clocks an die MMC/SD-Karte
		{
		mmc_write_byte(0xff);
		}
	
	//Sendet Commando CMD0 an MMC/SD-Karte
	uint8_t CMD[] = {0x40,0x00,0x00,0x00,0x00,0x95};
	while(mmc_write_command (CMD) !=1)
	{
		if (Timeout++ > 200)
			{
			MMC_Disable();
			return(1); //Abbruch bei Commando1 (Return Code1)
			}
	}
	//Sendet Commando CMD1 an MMC/SD-Karte
	Timeout = 0;
	CMD[0] = 0x41;//Commando 1
	CMD[5] = 0xFF;
	while( mmc_write_command (CMD) !=0)
	{
		if (Timeout++ > 400)
			{
			MMC_Disable();
			return(2); //Abbruch bei Commando2 (Return Code2)
			}
	}
	#if SPI_Mode
		//SPI Bus auf max Geschwindigkeit
		SPCR &= ~((1<<SPR0) | (1<<SPR1));
		SPSR = SPSR|(1<<SPI2X);
	#endif
	
	//set MMC_Chip_Select to high (MMC/SD-Karte Inaktiv)
	MMC_Disable();
	return(0);
}

// Neue Shutdownfunktion fr komplettes Abschalten der SD-Versorgung
void mmc_shutdown()
{
  SPCR = 0;
  MMC_Enable();						// CS low
  FETPORT |= (1 << SD_FET);			// SD Karte stromlos
  
  PRR |= (1 << PRSPI);			// SPI deaktivieren
  
}

//############################################################################
//Sendet ein Commando an die MMC/SD-Karte
static uint8_t mmc_write_command (uint8_t *cmd)
//############################################################################
{
	uint8_t tmp = 0xff;
	uint16_t Timeout = 0;

	//set MMC_Chip_Select to high (MMC/SD-Karte Inaktiv) 
	MMC_Disable();

	//sendet 8 Clock Impulse
	mmc_write_byte(0xFF);

	//set MMC_Chip_Select to low (MMC/SD-Karte Aktiv)
	MMC_Enable();

	//sendet 6 Byte Commando
	for (uint8_t a = 0;a<0x06;a++) //sendet 6 Byte Commando zur MMC/SD-Karte
		{
		mmc_write_byte(*cmd++);
		}

	//Wartet auf ein gltige Antwort von der MMC/SD-Karte
	while (tmp == 0xff)	
		{
		tmp = mmc_read_byte();
		if (Timeout++ > 500)
			{
			break; //Abbruch da die MMC/SD-Karte nicht Antwortet
			}
		}
	return(tmp);
}

//############################################################################
//Routine zum Empfangen eines Bytes von der MMC-Karte 
static uint8_t mmc_read_byte (void)
//############################################################################
{
	uint8_t Byte = 0;
#if SPI_Mode	//Routine fr Hardware SPI
	SPDR = 0xff;
	while(!(SPSR & (1<<SPIF))){};
	Byte = SPDR;

#else			//Routine fr Software SPI
	for (unsigned char a=8; a>0; a--) //das Byte wird Bitweise nacheinander Empangen MSB First
		{
		MMC_Write &=~(1<<SPI_Clock); //erzeugt ein Clock Impuls (Low) 
		
		if (bit_is_set(MMC_Read,SPI_DI) > 0) //Lesen des Pegels von MMC_DI
				{
				Byte |= (1<<(a-1));
				}
			else
				{
				Byte &=~(1<<(a-1));
				}
		MMC_Write |=(1<<SPI_Clock); //setzt Clock Impuls wieder auf (High)		
		}
#endif
	return (Byte);
}


//############################################################################
//Routine zum Senden eines Bytes zur MMC-Karte
static void mmc_write_byte (uint8_t Byte)
//############################################################################
{
#if SPI_Mode		//Routine fr Hardware SPI
	SPDR = Byte; 	//Sendet ein Byte
	while(!(SPSR & (1<<SPIF))) //Wartet bis Byte gesendet wurde
	{
	}
#else			//Routine fr Software SPI
	for (unsigned char a=8; a>0; a--) //das Byte wird Bitweise nacheinander Gesendet MSB First
	{
		if (bit_is_set(Byte,(a-1))>0)	//Ist Bit a in Byte gesetzt
				{
				MMC_Write |= (1<<SPI_DO); //Set Output High
				}
			else
				{
				MMC_Write &= ~(1<<SPI_DO); //Set Output Low
				}	
		MMC_Write &= ~(1<<SPI_Clock);	//erzeugt ein Clock Impuls (LOW)
		
		MMC_Write |= (1<<SPI_Clock); //setzt Clock Impuls wieder auf (High)
	}
	MMC_Write |= (1<<SPI_DO);	//setzt Output wieder auf High	
#endif
}

//############################################################################
//Routine zum schreiben eines Blocks(512Byte) auf die MMC/SD-Karte
uint8_t mmc_write_sector (uint32_t addr,uint8_t *Buffer)
//############################################################################
{
	uint8_t tmp;
	//Commando 24 zum schreiben eines Blocks auf die MMC/SD - Karte
	uint8_t cmd[] = {0x58,0x00,0x00,0x00,0x00,0xFF}; 
	
	/*Die Adressierung der MMC/SD-Karte wird in Bytes angegeben,
	  addr wird von Blocks zu Bytes umgerechnet danach werden 
	  diese in das Commando eingefgt*/
	  
	addr = addr << 9; //addr = addr * 512
	
	cmd[1] = ((addr & 0xFF000000) >>24 );
	cmd[2] = ((addr & 0x00FF0000) >>16 );
	cmd[3] = ((addr & 0x0000FF00) >>8 );
	//Sendet Commando cmd24 an MMC/SD-Karte (Write 1 Block/512 Bytes)
	tmp = mmc_write_command (cmd);
	if (tmp != 0)
		{
		return(tmp);
		}
			
	//Wartet einen Moment und sendet einen Clock an die MMC/SD-Karte
	for (uint8_t a=0;a<100;a++)
		{
		mmc_read_byte();
		}
	//Sendet Start Byte an MMC/SD-Karte
	mmc_write_byte(0xFE);	
	
	//Schreiben des Bolcks (512Bytes) auf MMC/SD-Karte
	for (uint16_t a=0;a<512;a++)
		{
	   mmc_write_byte(*Buffer++);
		}
	
	//CRC-Byte schreiben
	mmc_write_byte(0xFF); //Schreibt Dummy CRC
	mmc_write_byte(0xFF); //CRC Code wird nicht benutzt
	
	//Fehler beim schreiben? (Data Response XXX00101 = OK)
	if((mmc_read_byte()&0x1F) != 0x05) return(1);

	//Wartet auf MMC/SD-Karte Bussy
	while (mmc_read_byte() != 0xff){};
	
	//set MMC_Chip_Select to high (MMC/SD-Karte Inaktiv)
	MMC_Disable();
	
return(0);
}
/*
zeroes the whole SD Card
*/
void mmc_format()
{
  uint8_t cmd[6];
  uint16_t c2;
  led = 0;
  for(c2=0; c2<=7579; c2++)
  {
  led++;			// Bit 7 cycles about 30 times before finish. My transcend 512 MB MicroSD does ~300 kByte/s here
  					// that is about 27 minutes per 512 MB card.
  cmd[0] = 0x77;
  
  mmc_write_command(cmd);
  cmd[0] = 0x57;
  cmd[1] = 0;
  cmd[2] = 0;
  cmd[3] = 0;
  cmd[4] = 128;
  while(mmc_read_byte()!=0xff){};
  mmc_write_command(cmd);
  cmd[0] = 0x59;
  cmd[1] = c2 >> 8;
  cmd[2] = c2;
  cmd[3] = 0;
  cmd[4] = 0;
  while(mmc_read_byte()!=0xff){};
  mmc_write_command(cmd);
  // send 15 blocks with token 0xFC and one with 0xFD
  uint8_t count = 128;
  
  uint8_t a;
 
  while(count--)
  {
    while(mmc_read_byte()!=0xff){};
	mmc_write_byte(0xFC);
	a = 0;
	do
	{
	  mmc_write_byte(0);
	  mmc_write_byte(0);
    } while(--a);
	mmc_write_byte(0xFF);
	mmc_write_byte(0xFF);
    if((mmc_read_byte()&0x1F) !=0x05) return;
  }
  while(mmc_read_byte()!=0xFF){};
  mmc_write_byte(0xFD);
  }
  MMC_Disable();
}
/* */

//############################################################################
// Routine um angefangene Sektor-Read Operationen zu beenden
// changed/new by Matthias Larisch
void mmc_read_sector_finish(uint8_t *Buffer, uint8_t discard_bytes)
//############################################################################
{

	uint16_t temp = mmc_sector_remaining;
	mmc_sector_remaining = 0;

	while(temp--)
	{
		if(discard_bytes) mmc_read_byte();
		else *Buffer++ = mmc_read_byte();
	}
	//CRC-Byte auslesen
	mmc_read_byte();//CRC - Byte wird nicht ausgewertet
	mmc_read_byte();//CRC - Byte wird nicht ausgewertet
	
	//set MMC_Chip_Select to high (MMC/SD-Karte Inaktiv)
	MMC_Disable();
	
	return;
}

//############################################################################
//Routine zum Lesen eines Blocks(512Byte) von der MMC/SD-Karte
// startet nur den Lesevorgang, ohne Bytes zu lesen
// changed by Matthias Larisch
void mmc_read_sector_start (uint32_t addr)
//############################################################################
{	
	if(mmc_sector_remaining > 0)
	  mmc_read_sector_finish(0,1);		// finish last read sector
	//Commando 16 zum lesen eines Blocks von der MMC/SD - Karte
	uint8_t cmd[] = {0x51,0x00,0x00,0x00,0x00,0xFF}; 
	
	/*Die Adressierung der MMC/SD-Karte wird in Bytes angegeben,
	  addr wird von Blocks zu Bytes umgerechnet danach werden 
	  diese in das Commando eingefgt*/
	addr = addr << 9; //addr = addr * 512

	cmd[1] = ((addr & 0xFF000000) >>24 );
	cmd[2] = ((addr & 0x00FF0000) >>16 );
	cmd[3] = ((addr & 0x0000FF00) >>8 );

	if (mmc_write_command (cmd) != 0)
      return;

    uint8_t temp;
    //mmc_sector_remaining = 512;
	/* mmc_sector_remaining will always be zero when we are here so writing the
	   higher byte to 2 is enough. safes 6 bytes */
	asm volatile ("ldi %0, 2"	"\n\t"
	              "sts mmc_sector_remaining+1, %0"
				   :"=&d" (temp));

	while (mmc_read_byte() != 0xfe);

	return;
}

/* changed by Matthias Larisch */
// Diese Funktion liest count Bytes in den Buffer. Count |= 0x8000 fhrt
// dabei zum berlesen von count Bytes, ohne sie in den Buffer zu schreiben
/*uint8_t mmc_read_sector_continue(uint8_t *Buffer, uint16_t count)
{
  uint8_t temp = 0;
  uint8_t temp2;
  if(count & 0x8000)		// hack around: discard read bytes
  {
    temp = 1;
	count &= 0x01FF;		// 512 is max
  }
  if(count == 0) return 0;
  mmc_sector_remaining -= count;
  uint8_t retval = mmc_read_byte();
  if(count == 1) return retval;
  if(!temp) *Buffer++ = retval;
  while(--count)
  {
    temp2 = mmc_read_byte();
    if(!temp) *Buffer++ = temp2;
  }
  return 0;
}
*/

uint8_t mmc_read_sector_continue(uint8_t *Buffer, uint16_t count)
{
  uint8_t temp = 0;
  uint8_t temp2;
  if(count & 0x8000)		// hack around: discard read bytes
  {
    temp = 1;
	count &= 0x01FF;		// 512 is max
  }
  if(count == 0) return 0;
  mmc_sector_remaining -= count;
  //uint8_t retval = mmc_read_byte();
  //if(count == 1) return retval;
  //if(!temp) *Buffer++ = retval;
  while(count--)
  {
    temp2 = mmc_read_byte();
    if(!temp) *Buffer++ = temp2;
  }
  return 0;
}
