;-----------------------------------------------------------------------------;
; Fast integer squareroot routines for avr-gcc project          (C)ChaN, 2008
;-----------------------------------------------------------------------------;
; uint16_t isqrt32 (uint32_t n);
;-----------------------------------------------------------------------------:
; 32bit integer squareroot
;-----------------------------------------------------------------------------;
;   uint16_t isqrt32 (
;     uint32_t n
;   );
;
; Return Value:
;   Squareroot of n.
;
; Size  = 53 words
; Clock = 532..548 cycles
; Stack = 0 byte

.global isqrt32
.func isqrt32

isqrt32:
	clr	r0
	clr	r18
	clr	r19
	clr	r20
	ldi	r21, 1
	clr	r27
	clr	r30
	clr	r31
	ldi	r26, 16
1:	lsl	r22
	rol	r23
	rol	r24
	rol	r25
	rol	r0
	rol	r18
	rol	r19
	rol	r20
	lsl	r22
	rol	r23
	rol	r24
	rol	r25
	rol	r0
	rol	r18
	rol	r19
	rol	r20
	brpl	2f
	add	r0, r21
	adc	r18, r27
	adc	r19, r30
	adc	r20, r31
	rjmp	3f
2:	sub	r0, r21
	sbc	r18, r27
	sbc	r19, r30
	sbc	r20, r31
3:	lsl	r21
	rol	r27
	rol	r30
	andi	r21, 0b11111000
	ori	r21, 0b00000101
	sbrc	r20, 7
	subi	r21, 2
	dec	r26
	brne	1b
	lsr	r30
	ror	r27
	ror	r21
	lsr	r30
	ror	r27
	ror	r21
	mov	r24, r21
	mov	r25, r27
	ret
.endfunc


;-----------------------------------------------------------------------------:
; 32bit/32bit Unsigned Division
;
; Register Variables
;  Call:  var1[3:0] = dividend (0x00000000..0xffffffff) r25 24 23 22
;         var2[3:0] = divisor (0x00000001..0x7fffffff) r21 20 19 18
;         mod[3:0]  = <don't care> 31 30 r1 r0
;         lc        = <don't care> (high register must be allocated) 26
;
;  Result:var1[3:0] = var1[3:0] / var2[3:0]
;         var2[3:0] = <not changed>
;         mod[3:0]  = var1[3:0] % var2[3:0]
;         lc        = 0
;
; Size  = 26 words
; Clock = 549..677 cycles (+ret)
; Stack = 0 bytes
;
; taken from elm-chan.org, register reassignment august 2007 by Matthias Larisch
; for use as avr-gcc function

.global div32u
.func div32u

div32u:	
		;push r16
		clr	r0		;initialize variables
		;clr r1		;  mod = 0;
		clr	r30		;  lc = 32;
		clr	r31		;
		ldi	r26,32		;/
					;---- calcurating loop
	1:	lsl	r22		;var1 = var1 << 1;
		rol	r23		;
		rol	r24		;
		rol	r25		;/
		rol	r0		;mod = mod << 1 + carry;
		rol	r1		;
		rol	r30		;
		rol	r31		;/
		cp	r0,r18	;if (mod => var2) {
		cpc	r1,r19	; mod -= var2; var1++;
		cpc	r30,r20	; }
		cpc	r31,r21	;
		brcs	2f		;
		inc	r22		;
		sub	r0,r18	;
		sbc	r1,r19	;
		sbc	r30,r20	;
		sbc	r31,r21	;/
	2:	dec	r26		;if (--lc > 0)
		brne	1b		; continue loop;
		;pop r16
		clr r1
		ret

.endfunc


;-----------------------------------------------------------------------------:
; 16bit/16bit Unsigned Division
;
; Register Variables
;  Call:  var1[1:0] = dividend (0x0000..0xffff) r24 r25
;         var2[1:0] = divisor (0x0001..0x7fff) r22 r23
;         mod[1:0]  = <don't care> r21 r20
;         lc        = <don't care> (high register must be allocated) r19
;
;  Result:var1[1:0] = var1[1:0] / var2[1:0]
;         var2[1:0] = <not changed>
;         mod[1:0]  = var1[1:0] % var2[1:0]
;         lc        = 0
;
; Size  = 16 words
; Clock = 179..211 cycles (+ret)
; Stack = 0 byte
;
; taken from elm-chan.org. register reassignment for use as avr-gcc function
; august 2007 by Matthias Larisch
.global div16u
.func div16u

div16u:	
		clr	r20		;initialize variables
		clr	r21		;  mod = 0;
		ldi	r19,16		;  lc = 16;
					;---- calcurating loop
	1:	lsl	r24		;var1 = var1 << 1;
		rol	r25		;/
		rol	r20		;mod = mod << 1 + carry;
		rol	r21		;/
		cp	r20,r22	;if (mod => var2) {
		cpc	r21,r23	; mod -= var2; var1++;
		brcs	2f	; }
		inc	r24		;
		sub	r20,r22	;
		sbc	r21,r23	;/
	2:	dec	r19		;if (--lc > 0)
		brne	1b		; continue loop;
		ret
.endfunc

; 32 bit * 16 bit -> 32 bit multiplication
; input: r25 r24 r23 r22  factor 1
; input: r21 r20		  factor 2
; output:r25 r24 r23 r22  result
; uses: r31 r30 r27 r26 for result calculation
; clear r1 afterwards
;
; august 2007 by Matthias Larisch inspired by original AVR-GCC multiplication.
.global mul32_16u
.func mul32_16u

mul32_16u:
		mul r22, r20		; F10 * F20
		movw r26, r0
		mul r23, r21		; F11 * F21
		movw r30, r0
		mul r24, r20		; F12 * F20
		add r30, r0
		adc r31, r1
		mul r24, r21		; F12 * F21
		add r31, r0
		mul r25, r20		; F13 * F20
		add r31, r0
		clr r25
		mul r22, r21		; F10 * F21
		add r27, r0
		adc r30, r1
		adc r31, r25
		mul r23, r20		; F11 * F20
		add r27, r0
		adc r30, r1
		adc r31, r25
		movw r22, r26
		movw r24, r30
		clr r1
		ret

;F10: F20 *F21
;F11: F21 *F20
;F12: F20 *F21
;F13: F20 [F21]
