/*
  GPS Mini Navigator    for navigating predefined ways & extended gps logging

  Copyright (C) 2008  Matthias Larisch

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation; either version 2 of the License, or (at your option) any later
  version.
  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  You should have received a copy of the GNU General Public License along with
  this program; if not, see <http://www.gnu.org/licenses/>.

  File description: command interpreter for serial communication
*/

#include "egl.h"
#include "suart_commands.h"
#include "gps_uart.h"
#include "suart.h"
#include "filesystem_simple.h"
#include "mmc.h"

void suart_receive( void )
{
  uint8_t temp;
  static uint8_t command = 0;
  static union
  {
    uint8_t b[4];
	uint16_t w[2];
	uint32_t q;
  } command_arg;
  if(srx_done)	
  {
    temp = sgetchar();
	if(command == 0)
	{
	  command = temp;
	  command_arg.q = 0;
    }
	switch(command)
	{
      case SUART_GPS_FORWARD:
	    sputchar(SUART_SUCCESS);
        gps_enable(0);
		while(power_status.sw_uart_on)
		{
		  if(srx_done) gps_transmit(sgetchar());		// receive byte and output it
		  if(UCSR0A & (1 << RXC0)) sputchar(UDR0);		// and the other way too
        }
		gps_disable();
		command = 0;
        break;

      case SUART_CALL_BOOTLOADER:
	    // here we should finish all SD card crap
        /*cli();
		wdr();
		MCUSR &= ~(1 << WDRF);
        WDTCSR = (1 << WDE) | (1 << WDCE);
		WDTCSR = (1 << WDE) | (1 << WDP2) | (1 << WDP0);	// Enable Watchdog, 0.5 seconds
        // at this point we have 0.5 seconds until reset!
		sei();		// do we really need this? maybe we just wait in a loop here
		*/
		//wdt_reset();
		wdt_enable(WDTO_500MS);
		while(1){asm("nop");}		// loop here, else we will call this again and again... and nothing happens
		break;

      case SUART_CREATE_FILE:
	    if(command_arg.b[0] == 0)
		{
		  sputchar(SUART_SUCCESS);
        }
		else if(command_arg.b[0] == 1)
		{
		  sputchar(create_file(temp));
		  sputchar(SUART_COMPLETED);
		  command = 0;
        }
		command_arg.b[0]++;
        break;

      case SUART_DELETE_FILE:
	    if(command_arg.b[0] == 0)
		{
		  sputchar(SUART_SUCCESS);
        }
		else if(command_arg.b[0] == 1)
		{
		  delete_file(temp);
		  sputchar(SUART_COMPLETED);
		  command = 0;
        }
		command_arg.b[0]++;
		break;

      case SUART_WRITE_FILE:
	    if(command_arg.b[0] == 0)
		{
		  sputchar(SUART_SUCCESS);
		  command_arg.b[0]++;
        }
		else if(command_arg.b[0] == 1)
		{
		  command_arg.b[2] = temp;
		  command_arg.b[0]++;
        }
		else if(command_arg.b[0] == 2)
		{
		  command_arg.b[3] = temp;
		  sputchar(SUART_SUCCESS);
		  command_arg.b[0]++;
		  TCCR0B = 0;				// temporary disable of Timer 1
        }
		else
		{
		  temp = write_file(&temp,1);
		  if(temp == F_NEW_SECTOR)
		  {
		    sputchar(SUART_NEXT_SECTOR);
          }
		  else if(temp!=F_OK)
		  {
		    sputchar(SUART_MEDIUM_ERR);
			command = 0;
			TCCR0B = (1 << CS00) | (1 << CS01);		// reenable Timer 1
          }
          if(--command_arg.w[1] == 0)
		  {
		    sputchar(SUART_COMPLETED);
			close_write_file();
			command = 0;
			TCCR0B = (1 << CS00) | (1 << CS01);		// reenable Timer 1
          }
        }		  
		break;

      case SUART_READ_FILE:
	    if(command_arg.b[0] == 0)
		{
		  sputchar(SUART_SUCCESS);
		  command_arg.b[0]++;
        }
		else if(command_arg.b[0] == 1)
		{
		  sputchar(open_file(temp));
		  command_arg.b[0]++;
        }
		else if(command_arg.b[0] == 2)
		{
		  if(temp == SUART_STOP_COMMAND)
		  {
		    command = 0;
			sputchar(SUART_SUCCESS);
			close_read_file();
			break;
          }
		  if(read_file(buffer.b,507)!=F_OK)				// transmit 507 bytes at once (one sector of data)
		  												// this number is a workaround for having old data in the buffer
														// because read_file returns without zeroing the rest of the bytes after EOF
		  {
		    command_arg.b[0]++;
          }
          sputchar(SUART_CLUSTER_FOLLOWING);
		  /*for(uint8_t temp = 0; temp < 254; temp++)
		  {
		    sputchar(buffer.b[temp*2]);
			sputchar(buffer.b[temp*2+1]);
          }
		  */
		  for(uint16_t temp = 0; temp < 507; temp++)
		    sputchar(buffer.b[temp]);
        }
		else if(command_arg.b[0] == 3)
		{
		  command = 0;
		  close_read_file();
		  sputchar(SUART_FILE_FINISHED);
        }
        break;

      case SUART_MMC_INIT:
	    sputchar(mmc_init());
		command = 0;
		break;
      case SUART_MMC_FORMAT:
	    suart_deinit();									// leds indicate process
	    mmc_format();
		command = 0;
	  default:
	    command = 0;
        sputchar(SUART_BAD_COMMAND);
		break;
    }
  }
}

