/*
  GPS Mini Navigator    for navigating predefined ways & extended gps logging

  Copyright (C) 2008  Matthias Larisch

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation; either version 2 of the License, or (at your option) any later
  version.
  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  You should have received a copy of the GNU General Public License along with
  this program; if not, see <http://www.gnu.org/licenses/>.
  

  file description: LED/Key-UI implementation and some special functions

*/



#include "egl.h"
#include "gps_uart.h"
#include "mmc.h"
#include "filesystem_simple.h"
#include "keys.h"


uint8_t waypoint_options;
uint8_t menu;

uint8_t get_waypoint()
{
  uint16_t temp = navi_entry;
  uint8_t retval = 0;
  if(open_file(navi_file)==FT_WAY)
  {
    while (temp--)
	  retval = read_file((uint8_t *)&nav_to, 8);
    
	//if(nav_to.lat == 0 && nav_to.lon == 0) retval = 1;
	uint8_t counter = 0;
	for(uint8_t templ = 0; templ < 8; templ++)
	  if(nav_to.coords[templ] == 0) counter++;
    if(counter == 7) retval = 1;
  }
  else retval = 1;
  close_read_file();
  return retval;
}


const uint8_t menusize[] PROGMEM = {5, 3, 5, 3, 2};
void ui_do( void )
{
  //static uint8_t menu = 0;				// bit 7: menu active
  static uint8_t submenu = 0;
  static uint8_t called_submenu = 0;
  uint8_t led_t = 0;
  uint8_t led_cval = 0;			
  if(get_key_long (1 << KEY1))
  {
    menu &= 0x7F;	// goto menu
	submenu = 0;    // clear submenu
	status &= ~(1 << st_decimal_display);
	called_submenu = 0;
  }

  if((menu & 0x80) == 0)
  {	  
	if(get_key_short(1 << KEY0)) menu--;
	if(get_key_short(1 << KEY1)) menu++;
    if(get_key_long (1 << KEY0)) menu|= 0x80;	// show submenu

	if(menu == 0xFF) menu = 4;
	if(menu == 5) menu = 0;
  }
  else // in any menu point
  {
    uint8_t temp = 0;
    if(get_key_short(1 << KEY0))
	{
	  submenu--;
	  temp = 1;
    }
	if(get_key_short(1 << KEY1))
	{
	  submenu++;
	  temp = 1;
    }
	if (temp)
	{
	  status &= ~(1 << st_decimal_display);
	  called_submenu = 0;
    }
	temp = pgm_read_byte(&menusize[menu & 0x0F]);
	if(submenu == temp) submenu = 0;
	if(submenu == 0xFF) submenu = temp - 1;
    if(get_key_long(1 << KEY0))
	{
	  if(menu == 0x80)					// GPS
	  {
	    if(submenu == 0)				// LAT
		{
		  decimal_display(gps_log_entry.lat);
		}
		else if(submenu == 1)			// LON
		{
		  decimal_display(gps_log_entry.lon);
		}
		else if(submenu == 2)			// ALT
		{
		  decimal_display(gps_log_entry.alt);
		}
		else if(submenu == 3)			// SOG
		{
		}
		else if(submenu == 4)			// COG
		{
		}
	  }
	  else if(menu == 0x81)			// GPS LOG
	  {
	    if(submenu == 0)  				// Start/Stop
		{
		  if(status & (1 << st_gps_log))
		  {
			status &= ~(1 << st_gps_log);
			close_write_file();
		  }
		  else
		  {
		    status |= (1 << st_gps_log);
			led_cval = create_file(FT_LOG);
          }
		}
		else if(submenu == 1)			// TIME
		{
		  gps_log_interval++;
		  if(gps_log_interval >= GPS_LOG_INTERVALS) gps_log_interval = 0;
		}
		else if(submenu == 2)			// distance
		{
		  gps_log_distance++;
		  if(gps_log_distance >= GPS_LOG_DISTANCES) gps_log_distance = 0;
		}
      }
	  else if(menu == 0x82)			// WAYPOINTS
	  {
	    if(called_submenu)
		{
		  if(submenu == 0)				// Open a Waypoint / Waylist
		  {
		    while(open_file(++navi_file)!= FT_WAY)
			  if(navi_file == 0xFF) break;

			navi_entry = 1;
			if(waypoint_options & 2)		// reverse -> start with last waypoint
			{
			  while(get_waypoint() == 0)
			    navi_entry++;
              navi_entry--;
            }
			get_waypoint();
		  }		
		  else if(submenu == 1 || submenu == 2)		// go to next waypoint (submenu 1/2)
		  {
            if(navi_entry)
		    {
		      //gps_coords m2s2temp_coords;
			  //navi_entry+=(waypoint_options & 2) ? -1 : 1;
			  if(waypoint_options & 2) navi_entry--;
			  else navi_entry++;
			  get_waypoint();
            }
          }
        }
		if(submenu == 3)								// search next nearest waypoint
		{
		  if(navi_entry)
		  {
		    uint16_t lowest_distance = 0xFFFF;
			uint16_t lowest_distance_entry = 1;
			uint16_t temp_distance;
			while(nav_to.lat != 0)
			{
			  get_waypoint();
			  temp_distance = calculate_distance((gps_coords *)&gps_log_entry.coords, &nav_to);
			  if(temp_distance < lowest_distance)
			  {
			    lowest_distance = temp_distance;
				lowest_distance_entry = navi_entry;
              }
			  navi_entry++;
            }
			navi_entry = lowest_distance_entry;
			get_waypoint();
          }  
        }
		else if(submenu == 4)							// Options
		{
		  waypoint_options ++;								// Bit 0: automatically go to next waypoint when tracking waylist
		  if(waypoint_options >= 4) waypoint_options = 0;		// Bit 1: waylist reversion
		}    
	  }
	  else if(menu == 0x83)			// Create Waypoints
	  {
	    if(submenu == 0)			// create new waypoint file
		{
		  swap_write_files();			// gps log -> temp write
		  led_cval = create_file(FT_WAY);	
		  swap_write_files();			// waypoints -> temp write, gps log -> main write
		}
		else if(submenu == 1)		// warning: every waypoint need 2 sector writes on SD.
		{
		  swap_write_files();
		  write_file((uint8_t *)&gps_log_entry.coords, 8);		// write lat & lon (first 8 bytes in log entry)
		  swap_write_files();			
		}
		else if(submenu == 2)		// finish waypoint file
		{
		  swap_write_files();
		  close_write_file();
		  swap_write_files();
		}
      }
	  else if(menu == 0x84)			// System
	  {
	    if(submenu == 0)				// Shutdown/Standby (lets GPS Log running!)
		{
		  shutdown();
		}
		else if(submenu == 1)			// gps power
		{
		  if(FETPORT & (1 << GPS_FET)) gps_enable(1);
		  else gps_disable();
		}
	  }
	  called_submenu = 1;
    }
  }
  //if(menu == 0x83 && submenu == 2) led_t = power_status.GPS_FET_PWR << 7;
  /* following output is displayed together with menu/submenu indicator. leds 7 6 and 5 can be used */

  if(menu == 0x84 && submenu == 1)
  {
    led_t = (FETPORT & (1 << GPS_FET)) ? 0 : 128;		// gps powered?
	uint8_t temp = get_power();
	if(temp < 70) led_t |= 64;
	else if(temp < 105) led_t |= 32;
  }
  if(menu == 0x80) led_t |= (gps_log_entry.sog << 5) & 0xE0;	// gps fix state
  if(menu == 0x81 && submenu == 1)							// log time interval
    led_t = gps_log_interval << 5;
  if(menu == 0x81 && submenu == 2) 							// log distance interval
    led_t = gps_log_distance << 5;
  if(menu == 0x81 && submenu == 0) 							// gps log enabled?
	led_t = (status & (1 << st_gps_log)) << (7-st_gps_log);
  if(menu == 0x82 && submenu == 4)							// waypoint options
    led_t = waypoint_options << 6;
  
  if((status & (1 << st_decimal_display)) == 0) led_t |= (menu & 0x03) | ((submenu & 0x07) << 2);
  /* --------------------------------------------------------------------------------------------- */

  /* following output is displayed when a submenu is marked "called" (key 0 long press in submenu screen) */
  if(called_submenu)
  { 
    if(led_cval) led_t = led_cval;			// this overwrites some menu variables
    if(menu == 0x80)
    {
      if(submenu == 3)						// SOG
        led_t = div16u(gps_log_entry.sog,28);
      else if(submenu == 4)					// COG
	  {
	    led_t = display_course(current_gps_heading, 1);
      }
    }
	if(menu == 0x82)
	{
	  uint32_t temp32 = calculate_distance((gps_coords *)&gps_log_entry.coords, &nav_to);
	  uint8_t temp8 = temp32 >> 16;
	  uint8_t temp16 = temp32;
	  if((temp32 & 0x0000FFFF) < DISTANCE_AUTO_NEXT_WAYPOINT && (waypoint_options & 1))	// go to next waypoint
	  {
	    if(waypoint_options & 2) navi_entry--;
		else navi_entry++;
		get_waypoint();
      }
	  if(submenu == 0)
	    led_t = navi_file;								// displays currently opened waypoint/list file
	  else if(submenu == 1)								// displays distance current position <-> waypoint
	  {
		if(temp16 >=2048) temp16 = (temp16 >> 9) | 0x0080;		// led 7 indicates: distance display in 2048 meter steps
		else temp16 = temp16 >> 4;								// display distance in 16 meter steps
	    led_t = temp16;		
      }
      else if(submenu == 2)								// displays course from current position to waypoint
	    led_t = display_course(temp8,0);		// course
    }
  }
  /* ----------------------------------------------------------------------------------------------------- */
  led = led_t;
}
